# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'maskWidgetUI.ui'
##
## Created by: Qt User Interface Compiler version 6.10.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QFrame, QGridLayout,
    QHBoxLayout, QLabel, QPushButton, QRadioButton,
    QSizePolicy, QSlider, QSpacerItem, QSpinBox,
    QWidget)

class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName(u"Form")
        Form.resize(363, 366)
        self.horizontalLayout = QHBoxLayout(Form)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.toolsFrame = QFrame(Form)
        self.toolsFrame.setObjectName(u"toolsFrame")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.toolsFrame.sizePolicy().hasHeightForWidth())
        self.toolsFrame.setSizePolicy(sizePolicy)
        self.toolsFrame.setFrameShape(QFrame.StyledPanel)
        self.toolsFrame.setFrameShadow(QFrame.Raised)
        self.gridLayout = QGridLayout(self.toolsFrame)
        self.gridLayout.setObjectName(u"gridLayout")
        self.lineMaskButton = QPushButton(self.toolsFrame)
        self.lineMaskButton.setObjectName(u"lineMaskButton")

        self.gridLayout.addWidget(self.lineMaskButton, 3, 0, 1, 1)

        self.frame = QFrame(self.toolsFrame)
        self.frame.setObjectName(u"frame")
        self.frame.setMinimumSize(QSize(150, 150))
        self.frame.setFrameShape(QFrame.StyledPanel)
        self.frame.setFrameShadow(QFrame.Raised)
        self.gridLayout_2 = QGridLayout(self.frame)
        self.gridLayout_2.setSpacing(1)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.gridLayout_2.setContentsMargins(1, 1, 1, 1)
        self.horizontalSpacer_3 = QSpacerItem(88, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.gridLayout_2.addItem(self.horizontalSpacer_3, 2, 2, 1, 1)

        self.xMinBox = QSpinBox(self.frame)
        self.xMinBox.setObjectName(u"xMinBox")

        self.gridLayout_2.addWidget(self.xMinBox, 3, 2, 1, 1)

        self.horizontalSpacer_4 = QSpacerItem(88, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.gridLayout_2.addItem(self.horizontalSpacer_4, 3, 1, 1, 1)

        self.yMinBox = QSpinBox(self.frame)
        self.yMinBox.setObjectName(u"yMinBox")

        self.gridLayout_2.addWidget(self.yMinBox, 2, 1, 1, 1)

        self.xMaxBox = QSpinBox(self.frame)
        self.xMaxBox.setObjectName(u"xMaxBox")

        self.gridLayout_2.addWidget(self.xMaxBox, 1, 2, 1, 1)

        self.horizontalSpacer_5 = QSpacerItem(88, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.gridLayout_2.addItem(self.horizontalSpacer_5, 3, 3, 1, 1)

        self.horizontalSpacer = QSpacerItem(88, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.gridLayout_2.addItem(self.horizontalSpacer, 1, 1, 1, 1)

        self.horizontalSpacer_2 = QSpacerItem(88, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.gridLayout_2.addItem(self.horizontalSpacer_2, 1, 3, 1, 1)

        self.yMaxBox = QSpinBox(self.frame)
        self.yMaxBox.setObjectName(u"yMaxBox")

        self.gridLayout_2.addWidget(self.yMaxBox, 2, 3, 1, 1)

        self.label = QLabel(self.frame)
        self.label.setObjectName(u"label")

        self.gridLayout_2.addWidget(self.label, 4, 0, 1, 1)

        self.label_2 = QLabel(self.frame)
        self.label_2.setObjectName(u"label_2")

        self.gridLayout_2.addWidget(self.label_2, 0, 4, 1, 1)

        self.yMaxSlider = QSlider(self.frame)
        self.yMaxSlider.setObjectName(u"yMaxSlider")
        self.yMaxSlider.setOrientation(Qt.Vertical)

        self.gridLayout_2.addWidget(self.yMaxSlider, 1, 4, 3, 1)

        self.yMinSlider = QSlider(self.frame)
        self.yMinSlider.setObjectName(u"yMinSlider")
        self.yMinSlider.setOrientation(Qt.Vertical)

        self.gridLayout_2.addWidget(self.yMinSlider, 1, 0, 3, 1)

        self.xMaxSlider = QSlider(self.frame)
        self.xMaxSlider.setObjectName(u"xMaxSlider")
        self.xMaxSlider.setOrientation(Qt.Horizontal)

        self.gridLayout_2.addWidget(self.xMaxSlider, 0, 1, 1, 3)

        self.xMinSlider = QSlider(self.frame)
        self.xMinSlider.setObjectName(u"xMinSlider")
        self.xMinSlider.setOrientation(Qt.Horizontal)

        self.gridLayout_2.addWidget(self.xMinSlider, 4, 1, 1, 3)


        self.gridLayout.addWidget(self.frame, 4, 0, 1, 2)

        self.ellipseMaskButton = QPushButton(self.toolsFrame)
        self.ellipseMaskButton.setObjectName(u"ellipseMaskButton")

        self.gridLayout.addWidget(self.ellipseMaskButton, 2, 1, 1, 1)

        self.addROI = QRadioButton(self.toolsFrame)
        self.addROI.setObjectName(u"addROI")
        self.addROI.setChecked(True)

        self.gridLayout.addWidget(self.addROI, 1, 0, 1, 1)

        self.rectMaskButton = QPushButton(self.toolsFrame)
        self.rectMaskButton.setObjectName(u"rectMaskButton")

        self.gridLayout.addWidget(self.rectMaskButton, 2, 0, 1, 1)

        self.subtractROI = QRadioButton(self.toolsFrame)
        self.subtractROI.setObjectName(u"subtractROI")

        self.gridLayout.addWidget(self.subtractROI, 1, 1, 1, 1)

        self.polyMaskButton = QPushButton(self.toolsFrame)
        self.polyMaskButton.setObjectName(u"polyMaskButton")

        self.gridLayout.addWidget(self.polyMaskButton, 3, 1, 1, 1)

        self.archList = QComboBox(self.toolsFrame)
        self.archList.setObjectName(u"archList")

        self.gridLayout.addWidget(self.archList, 0, 0, 1, 1)

        self.clearButton = QPushButton(self.toolsFrame)
        self.clearButton.setObjectName(u"clearButton")

        self.gridLayout.addWidget(self.clearButton, 0, 1, 1, 1)

        self.setCurrent = QPushButton(self.toolsFrame)
        self.setCurrent.setObjectName(u"setCurrent")

        self.gridLayout.addWidget(self.setCurrent, 5, 0, 1, 1)

        self.setGlobal = QPushButton(self.toolsFrame)
        self.setGlobal.setObjectName(u"setGlobal")

        self.gridLayout.addWidget(self.setGlobal, 5, 1, 1, 1)


        self.horizontalLayout.addWidget(self.toolsFrame)


        self.retranslateUi(Form)
        self.xMaxBox.valueChanged.connect(self.xMaxSlider.setValue)
        self.xMaxSlider.sliderMoved.connect(self.xMaxBox.setValue)
        self.yMaxBox.valueChanged.connect(self.yMaxSlider.setValue)
        self.yMaxSlider.sliderMoved.connect(self.yMaxBox.setValue)
        self.xMinBox.valueChanged.connect(self.xMinSlider.setValue)
        self.xMinSlider.sliderMoved.connect(self.xMinBox.setValue)
        self.yMinSlider.sliderMoved.connect(self.yMinBox.setValue)
        self.yMinBox.valueChanged.connect(self.yMinSlider.setValue)

        QMetaObject.connectSlotsByName(Form)
    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", u"Form", None))
        self.lineMaskButton.setText(QCoreApplication.translate("Form", u"Line", None))
        self.label.setText(QCoreApplication.translate("Form", u"Min", None))
        self.label_2.setText(QCoreApplication.translate("Form", u"Max", None))
        self.ellipseMaskButton.setText(QCoreApplication.translate("Form", u"Ellipse", None))
        self.addROI.setText(QCoreApplication.translate("Form", u"Add", None))
        self.rectMaskButton.setText(QCoreApplication.translate("Form", u"Rect", None))
        self.subtractROI.setText(QCoreApplication.translate("Form", u"Subtract", None))
        self.polyMaskButton.setText(QCoreApplication.translate("Form", u"Poly", None))
        self.clearButton.setText(QCoreApplication.translate("Form", u"Clear", None))
        self.setCurrent.setText(QCoreApplication.translate("Form", u"Set Current", None))
        self.setGlobal.setText(QCoreApplication.translate("Form", u"Set Global", None))
    # retranslateUi




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'rangeWidgetUI.ui'
##
## Created by: Qt User Interface Compiler version 6.10.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QDoubleSpinBox, QFrame,
    QHBoxLayout, QLabel, QSizePolicy, QSpinBox,
    QVBoxLayout, QWidget)

class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName(u"Form")
        Form.resize(1440, 817)
        self.verticalLayout_2 = QVBoxLayout(Form)
        self.verticalLayout_2.setSpacing(0)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.verticalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.frame_3 = QFrame(Form)
        self.frame_3.setObjectName(u"frame_3")
        self.frame_3.setFrameShape(QFrame.StyledPanel)
        self.frame_3.setFrameShadow(QFrame.Raised)
        self.verticalLayout = QVBoxLayout(self.frame_3)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.titleLabel = QLabel(self.frame_3)
        self.titleLabel.setObjectName(u"titleLabel")
        font = QFont()
        font.setBold(True)
        self.titleLabel.setFont(font)
        self.titleLabel.setAlignment(Qt.AlignCenter)

        self.verticalLayout.addWidget(self.titleLabel)

        self.frame = QFrame(self.frame_3)
        self.frame.setObjectName(u"frame")
        self.frame.setFrameShape(QFrame.NoFrame)
        self.frame.setFrameShadow(QFrame.Raised)
        self.frame.setLineWidth(0)
        self.horizontalLayout = QHBoxLayout(self.frame)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.rangeLabel = QLabel(self.frame)
        self.rangeLabel.setObjectName(u"rangeLabel")
        self.rangeLabel.setFont(font)

        self.horizontalLayout.addWidget(self.rangeLabel)

        self.low = QDoubleSpinBox(self.frame)
        self.low.setObjectName(u"low")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.low.sizePolicy().hasHeightForWidth())
        self.low.setSizePolicy(sizePolicy)
        self.low.setMinimumSize(QSize(0, 0))
        self.low.setMaximumSize(QSize(68, 16777215))
        self.low.setMaximum(1000000.000000000000000)
        self.low.setSingleStep(0.100000000000000)
        self.low.setValue(0.000000000000000)

        self.horizontalLayout.addWidget(self.low)

        self.label_4 = QLabel(self.frame)
        self.label_4.setObjectName(u"label_4")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.label_4.sizePolicy().hasHeightForWidth())
        self.label_4.setSizePolicy(sizePolicy1)
        self.label_4.setMaximumSize(QSize(20, 16777215))
        self.label_4.setAlignment(Qt.AlignCenter)

        self.horizontalLayout.addWidget(self.label_4)

        self.high = QDoubleSpinBox(self.frame)
        self.high.setObjectName(u"high")
        sizePolicy.setHeightForWidth(self.high.sizePolicy().hasHeightForWidth())
        self.high.setSizePolicy(sizePolicy)
        self.high.setMaximumSize(QSize(68, 16777215))
        self.high.setSingleStep(0.100000000000000)

        self.horizontalLayout.addWidget(self.high)

        self.units = QComboBox(self.frame)
        self.units.setObjectName(u"units")
        self.units.setMaximumSize(QSize(68, 16777215))

        self.horizontalLayout.addWidget(self.units)


        self.verticalLayout.addWidget(self.frame)

        self.frame_2 = QFrame(self.frame_3)
        self.frame_2.setObjectName(u"frame_2")
        self.frame_2.setFrameShape(QFrame.NoFrame)
        self.frame_2.setFrameShadow(QFrame.Raised)
        self.frame_2.setLineWidth(0)
        self.horizontalLayout_2 = QHBoxLayout(self.frame_2)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(0, 0, 0, 0)
        self.pointsLabel = QLabel(self.frame_2)
        self.pointsLabel.setObjectName(u"pointsLabel")
        self.pointsLabel.setFont(font)

        self.horizontalLayout_2.addWidget(self.pointsLabel)

        self.points = QSpinBox(self.frame_2)
        self.points.setObjectName(u"points")
        sizePolicy.setHeightForWidth(self.points.sizePolicy().hasHeightForWidth())
        self.points.setSizePolicy(sizePolicy)
        self.points.setMaximumSize(QSize(68, 16777215))
        self.points.setMinimum(1)
        self.points.setMaximum(10000000)

        self.horizontalLayout_2.addWidget(self.points)

        self.stepLabel = QLabel(self.frame_2)
        self.stepLabel.setObjectName(u"stepLabel")
        self.stepLabel.setMaximumSize(QSize(60, 16777215))
        self.stepLabel.setFont(font)

        self.horizontalLayout_2.addWidget(self.stepLabel)

        self.step = QDoubleSpinBox(self.frame_2)
        self.step.setObjectName(u"step")
        sizePolicy.setHeightForWidth(self.step.sizePolicy().hasHeightForWidth())
        self.step.setSizePolicy(sizePolicy)
        self.step.setMaximumSize(QSize(68, 16777215))
        self.step.setDecimals(3)
        self.step.setMinimum(0.001000000000000)
        self.step.setMaximum(1000000.000000000000000)
        self.step.setSingleStep(0.100000000000000)

        self.horizontalLayout_2.addWidget(self.step)


        self.verticalLayout.addWidget(self.frame_2)


        self.verticalLayout_2.addWidget(self.frame_3)


        self.retranslateUi(Form)

        QMetaObject.connectSlotsByName(Form)
    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", u"Form", None))
        self.titleLabel.setText(QCoreApplication.translate("Form", u"Title", None))
        self.rangeLabel.setText(QCoreApplication.translate("Form", u"Range", None))
        self.label_4.setText(QCoreApplication.translate("Form", u"to", None))
        self.pointsLabel.setText(QCoreApplication.translate("Form", u"Points", None))
        self.stepLabel.setText(QCoreApplication.translate("Form", u"Step Size", None))
    # retranslateUi




# -*- coding: utf-8 -*-
"""
@author: thampy, walroth
"""

# Standard library imports
import logging
import os
import fnmatch
import numpy as np
from pathlib import Path

logger = logging.getLogger(__name__)

# Qt imports
from pyqtgraph.Qt import QtCore, QtWidgets
from pyqtgraph.parametertree import ParameterTree, Parameter

# Project imports
from ssrl_xrd_tools.core.containers import PONI
from ssrl_xrd_tools.io.metadata import read_image_metadata
from .wrangler_widget import wranglerWidget
from .spec_wrangler_thread import specThread, _get_scan_info  # noqa: F401
from .ui.specUI import Ui_Form
from ....gui_utils import NamedActionParameter
from xdart.utils import get_fname_dir, match_img_detector
from xdart.utils.session import load_session, save_session


QFileDialog = QtWidgets.QFileDialog
QDialog = QtWidgets.QDialog
QMessageBox = QtWidgets.QMessageBox
QPushButton = QtWidgets.QPushButton


def_poni_file = '' # '/Users/vthampy/SSRL_Data/RDA/static_det_test_data/test_xfc_data/test_xfc.poni'
def_img_file = '' # '/Users/vthampy/SSRL_Data/RDA/static_det_test_data/test_xfc_data/images/images_0005.tif'

if not os.path.exists(def_poni_file):
    def_poni_file = ''
    def_img_file = ''

params = [
    {'name': 'Calibration', 'type': 'group', 'children': [
        {'name': 'poni_file', 'title': 'PONI File    ', 'type': 'str', 'value': def_poni_file},
        NamedActionParameter(name='poni_file_browse', title='Browse...'),
    ], 'expanded': True},
    {'name': 'Signal', 'type': 'group', 'children': [
        {'name': 'inp_type', 'title': '', 'type': 'list',
         'values': ['Image Series', 'Image Directory', 'Single Image'], 'value': 'Image Series'},
        {'name': 'File', 'title': 'Image File   ', 'type': 'str', 'value': def_img_file},
        NamedActionParameter(name='img_file_browse', title='Browse...'),
        {'name': 'img_dir', 'title': 'Directory', 'type': 'str', 'value': '', 'visible': False},
        NamedActionParameter(name='img_dir_browse', title='Browse...', visible=False),
        {'name': 'include_subdir', 'title': 'Subdirectories', 'type': 'bool', 'value': False, 'visible': False},
        {'name': 'img_ext', 'title': 'File Type  ', 'type': 'list',
         'values': ['tif', 'raw', 'h5', 'mar3450'], 'value': 'tif', 'visible': False},
        {'name': 'series_average', 'title': 'Average Scan', 'type': 'bool', 'value': False, 'visible': True},
        {'name': 'meta_ext', 'title': 'Meta File', 'type': 'list',
         'values': ['None', 'txt', 'pdi', 'SPEC'], 'value': 'txt'},
        {'name': 'Filter', 'type': 'str', 'value': '', 'visible': False},
        {'name': 'write_mode', 'title': 'Write Mode  ', 'type': 'list',
         'values': ['Append', 'Overwrite'], 'value': 'Append'},
        {'name': 'mask_file', 'title': 'Mask File', 'type': 'str', 'value': ''},
        NamedActionParameter(name='mask_file_browse', title='Browse...'),
    ], 'expanded': True, 'visible': False},
    {'name': 'GI', 'title': 'Grazing Incidence', 'type': 'group', 'children': [
        {'name': 'Grazing', 'type': 'bool', 'value': False},
        {'name': 'th_motor', 'title': 'Theta Motor', 'type': 'list', 'values': ['th'],
         'value': 'th'},
        {'name': 'th_val', 'title': 'Theta', 'type': 'str', 'value': '0.1', 'visible': False},
        {'name': 'sample_orientation', 'title': 'Sample Orientation', 'type': 'int', 'value': 4,
         'limits': (1, 8), 'step': 1,
         'tip': 'EXIF sample orientation (1-8) for pyFAI FiberIntegrator'},
        {'name': 'tilt_angle', 'title': 'Tilt Angle', 'type': 'float', 'value': 0.0,
         'step': 0.1, 'tip': 'Chi offset / tilt angle in degrees for FiberIntegrator'},
        # gi_mode_1d / gi_mode_2d are controlled by the integrator panel
        # (axis1D / axis2D combos).  Kept here for session persistence only.
        {'name': 'gi_mode_1d', 'title': '1D Mode', 'type': 'list',
         'values': {'Q': 'q_total', u'Q\u1D62\u209A': 'q_ip', u'Q\u2092\u2092\u209A': 'q_oop'},
         'value': 'q_total', 'visible': False},
        {'name': 'gi_mode_2d', 'title': '2D Mode', 'type': 'list',
         'values': {u'Q\u1D62\u209A\u2013Q\u2092\u2092\u209A': 'qip_qoop', u'Q-\u03C7': 'q_chi'},
         'value': 'qip_qoop', 'visible': False},
    ], 'expanded': False, 'visible': False},
    {'name': 'Mask', 'title': 'Intensity Threshold', 'type': 'group', 'children': [
        {'name': 'Threshold', 'type': 'bool', 'value': False},
        {'name': 'min', 'title': 'Min', 'type': 'int', 'value': 0},
        {'name': 'max', 'title': 'Max', 'type': 'int', 'value': 0},
    ], 'expanded': False, 'visible': False},
    {'name': 'BG', 'title': 'Background', 'type': 'group', 'children': [
        {'name': 'bg_type', 'title': '', 'type': 'list',
         'values': ['None', 'Single BG File', 'Series Average', 'BG Directory'], 'value': 'None'},
        {'name': 'File', 'title': 'BG File', 'type': 'str', 'value': '', 'visible': False},
        NamedActionParameter(name='bg_file_browse', title='Browse...', visible=False),
        {'name': 'Match', 'title': 'Match Parameter', 'type': 'group', 'children': [
            {'name': 'Parameter', 'type': 'list', 'values': ['None'], 'value': 'None'},
            {'name': 'match_fname', 'title': 'Match File Root', 'type': 'bool', 'value': False},
            {'name': 'bg_dir', 'title': 'Directory', 'type': 'str', 'value': ''},
            NamedActionParameter(name='bg_dir_browse', title='Browse...'),
            {'name': 'Filter', 'type': 'str', 'value': ''},
        ], 'expanded': True, 'visible': False},
        {'name': 'Scale', 'type': 'float', 'value': 1, 'visible': False},
        {'name': 'norm_channel', 'title': 'Normalize', 'type': 'list', 'values': ['bstop'], 'value': 'bstop',
         'visible': False},
    ], 'expanded': False, 'visible': False},
    {'name': 'h5_dir', 'title': 'Save Path', 'type': 'str', 'value': get_fname_dir(), 'enabled': False},
    NamedActionParameter(name='h5_dir_browse', title='Browse...', visible=False),
]

ctr = 1



class specWrangler(wranglerWidget):
    """Widget for integrating data associated with spec file. Can be
    used "live", will continue to poll data folders until image data
    and corresponding spec data are available.

    attributes:
        command_queue: Queue, used to send commands to thread
        file_lock, mp.Condition, process safe lock for file access
        fname: str, path to data file
        parameters: pyqtgraph Parameter, stores parameters from user
        scan_name: str, current scan name, used to handle syncing data
        sphere_args: dict, used as **kwargs in sphere initialization.
            see EwaldSphere.
        thread: wranglerThread or subclass, QThread for controlling
            processes
        timeout: int, how long before thread stops looking for new
            data.
        tree: pyqtgraph ParameterTree, stores and organizes parameters
        ui: Ui_Form from qtdesigner

    methods:
        cont, pause, stop: functions to pass continue, pause, and stop
            commands to thread via command_queue
        enabled: Enables or disables interactivity
        set_image_dir: sets the image directory
        set_poni_file: sets the calibration poni file
        set_spec_file: sets the spec data file
        set_fname: Method to safely change file name
        setup: Syncs thread parameters prior to starting

    signals:
        finished: Connected to thread.finished signal
        sigStart: Tells tthetaWidget to start the thread and prepare
            for new data.
        sigUpdateData: int, signals a new arch has been added.
        sigUpdateFile: (str, str, bool, str, bool, bool), sends new scan_name, file name
            GI flag (grazing incidence), theta motor for GI, single_image and
            series_average flag to static_scan_Widget.
        sigUpdateGI: bool, signals the grazing incidence condition has changed.
        showLabel: str, connected to thread showLabel signal, sets text
            in specLabel
    """
    showLabel = QtCore.Signal(str)

    def __init__(self, fname, file_lock, sphere, data_1d, data_2d, parent=None):
        """fname: str, file path
        file_lock: mp.Condition, process safe lock
        """
        super().__init__(fname, file_lock, parent)

        # Scan Parameters
        self.poni = None
        self.scan_parameters = []
        self.counters = []
        self.motors = []
        self.command = None
        self.sphere = sphere
        self.data_1d = data_1d
        self.data_2d = data_2d

        # Setup gui elements
        self.ui = Ui_Form()
        self.ui.setupUi(self)
        self.ui.startButton.clicked.connect(self.start)
        # self.ui.startButton.clicked.connect(self.sigStart.emit)
        self.ui.pauseButton.clicked.connect(self.pause)
        self.ui.stopButton.clicked.connect(self.stop)
        self.ui.continueButton.clicked.connect(self.cont)
        self.ui.processingModeCombo.currentTextChanged.connect(self._on_mode_changed)
        self.ui.liveCheckBox.stateChanged.connect(self._on_mode_changed)
        self.ui.batchCheckBox.stateChanged.connect(self._on_mode_changed)
        self.ui.processingModeCombo.currentTextChanged.connect(lambda _: self._save_to_session())
        self.ui.liveCheckBox.stateChanged.connect(lambda _: self._save_to_session())
        self.ui.batchCheckBox.stateChanged.connect(lambda _: self._save_to_session())
        self._on_mode_changed()

        self.showLabel.connect(self.ui.specLabel.setText)

        # Setup parameter tree
        self.tree = ParameterTree()
        self.tree.setMinimumWidth(150)
        self.stylize_ParameterTree()
        self.parameters = Parameter.create(
            name='spec_wrangler', type='group', children=params
        )
        self.tree.setParameters(self.parameters, showTop=False)
        # Squeeze parameter tree columns to reduce panel width
        header = self.tree.header()
        header.setStretchLastSection(True)
        header.resizeSection(0, 130)  # name column
        self.layout = QtWidgets.QVBoxLayout(self.ui.paramFrame)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.addWidget(self.tree)

        # Set attributes from Parameter Tree and a couple more
        # Calibration
        self.poni_file = self.parameters.child('Calibration').child('poni_file').value()
        self.get_poni_dict()

        # Signal
        self.inp_type = self.parameters.child('Signal').child('inp_type').value()
        self.img_file = self.parameters.child('Signal').child('File').value()
        self.img_dir = self.parameters.child('Signal').child('img_dir').value()
        self.include_subdir = self.parameters.child('Signal').child('include_subdir').value()
        self.img_ext = self.parameters.child('Signal').child('img_ext').value()
        self.single_img = True if self.inp_type == 'Single Image' else False
        self.file_filter = self.parameters.child('Signal').child('Filter').value()
        self.series_average = self.parameters.child('Signal').child('series_average').value()
        self.meta_ext = self.parameters.child('Signal').child('meta_ext').value()
        if self.meta_ext == 'None':
            self.meta_ext = None

        # Mask
        self.mask_file = self.parameters.child('Signal').child('mask_file').value()

        # Threshold
        self.apply_threshold = self.parameters.child('Mask').child('Threshold').value()
        self.threshold_min = self.parameters.child('Mask').child('min').value()
        self.threshold_max = self.parameters.child('Mask').child('max').value()

        # Write Mode
        self.write_mode = self.parameters.child('Signal').child('write_mode').value()

        # Background
        self.bg_type = self.parameters.child('BG').child('bg_type').value()
        self.bg_file = self.parameters.child('BG').child('File').value()
        self.bg_dir = self.parameters.child('BG').child('Match').child('bg_dir').value()
        self.bg_matching_par = self.parameters.child('BG').child('Match').child('Parameter').value()
        self.bg_match_fname = self.parameters.child('BG').child('Match').child('match_fname').value()
        self.bg_file_filter = self.parameters.child('BG').child('Match').child('Filter').value()
        self.bg_scale = self.parameters.child('BG').child('Scale').value()
        self.bg_norm_channel = self.parameters.child('BG').child('norm_channel').value()

        # Grazing Incidence
        self.gi = self.parameters.child('GI').child('Grazing').value()
        self.th_mtr = self.parameters.child('GI').child('th_motor').value()
        self.sample_orientation = self.parameters.child('GI').child('sample_orientation').value()
        self.tilt_angle = self.parameters.child('GI').child('tilt_angle').value()
        # gi_mode_1d / gi_mode_2d are driven by the integrator panel;
        # default here, actual values set from sphere.bai_*_args at thread start.
        self.gi_mode_1d = self.sphere.bai_1d_args.get('gi_mode_1d', 'q_total')
        self.gi_mode_2d = self.sphere.bai_2d_args.get('gi_mode_2d', 'qip_qoop')

        # HDF5 Save Path
        self.h5_dir = self.parameters.child('h5_dir').value()

        # NOTE: Integration Advanced button (self.ui.advancedButton) is wired
        # in static_scan_widget.set_wrangler() to show the integratorTree's
        # existing advancedWidget1D / advancedWidget2D dialogs directly.

        # Wire signals from parameter tree based buttons
        self.parameters.sigTreeStateChanged.connect(self.setup)
        self.parameters.sigTreeStateChanged.connect(self._save_to_session)

        self.parameters.child('Calibration').child('poni_file_browse').sigActivated.connect(
            self.set_poni_file
        )
        self.parameters.child('Calibration').child('poni_file').sigValueChanged.connect(
            self.get_poni_dict
        )
        self.parameters.child('Signal').child('inp_type').sigValueChanged.connect(
            self.set_inp_type
        )
        self.parameters.child('Signal').child('img_file_browse').sigActivated.connect(
            self.set_img_file
        )
        self.parameters.child('Signal').child('img_dir_browse').sigActivated.connect(
            self.set_img_dir
        )
        self.parameters.child('Signal').child('mask_file_browse').sigActivated.connect(
            self.set_mask_file
        )
        self.parameters.child('Signal').child('series_average').sigValueChanged.connect(
            self.set_series_average
        )
        self.parameters.child('Signal').child('meta_ext').sigValueChanged.connect(
            self.set_meta_ext
        )
        self.parameters.child('BG').child('bg_type').sigValueChanged.connect(
            self.set_bg_type
        )
        self.parameters.child('BG').child('bg_file_browse').sigActivated.connect(
            self.set_bg_file
        )
        self.parameters.child('BG').child('Match').child('bg_dir_browse').sigActivated.connect(
            self.set_bg_dir
        )
        self.parameters.child('BG').child('Match').child('Parameter').sigValueChanged.connect(
            self.set_bg_matching_par
        )
        self.parameters.child('BG').child('norm_channel').sigValueChanged.connect(
            self.set_bg_norm_channel
        )
        self.parameters.child('GI').child('th_motor').sigValueChanged.connect(
            self.set_gi_th_motor
        )
        self.parameters.child('GI').child('th_val').sigValueChanged.connect(
            self.set_gi_th_motor
        )
        self.parameters.child('h5_dir_browse').sigActivated.connect(
            self.set_h5_dir
        )

        # Setup thread
        self.thread = specThread(
            self.command_queue,
            self.sphere_args,
            self.file_lock,
            self.fname,
            self.h5_dir,
            self.scan_name,
            self.single_img,
            self.poni,
            self.inp_type,
            self.img_file,
            self.img_dir,
            self.include_subdir,
            self.img_ext,
            self.series_average,
            self.meta_ext,
            self.file_filter,
            self.mask_file,
            self.write_mode,
            self.bg_type,
            self.bg_file,
            self.bg_dir,
            self.bg_matching_par,
            self.bg_match_fname,
            self.bg_file_filter,
            self.bg_scale,
            self.bg_norm_channel,
            self.gi,
            self.th_mtr,
            self.sample_orientation,
            self.tilt_angle,
            self.gi_mode_1d,
            self.gi_mode_2d,
            self.command,
            self.sphere,
            self.data_1d,
            self.data_2d,
            live_mode=self.live_mode,
            max_cores=self.ui.maxCoresSpinBox.value(),
            parent=self,
        )

        self.thread.showLabel.connect(self.ui.specLabel.setText)
        self.thread.sigUpdateFile.connect(self.sigUpdateFile.emit)
        self.thread.finished.connect(self.finished.emit)
        self.thread.sigUpdate.connect(self.sigUpdateData.emit)
        # self.thread.sigUpdateArch.connect(self.sigUpdateArch.emit)
        self.thread.sigUpdateGI.connect(self.sigUpdateGI.emit)

        # Enable/disable buttons initially
        self.ui.pauseButton.setEnabled(False)
        self.ui.continueButton.setEnabled(False)
        self.ui.stopButton.setEnabled(False)

        self.setup()
        self._restore_from_session()

    # --- Session persistence ---

    # Flat map: (session_key, param_path_tuple, is_path, self_attr)
    # is_path=True  → only restore if the path exists on disk
    # self_attr     → attribute to sync on restore (None = handled by sigValueChanged)
    _SESSION_PARAMS = [
        ('poni_file',      ('Calibration', 'poni_file'),             True,  'poni_file'),
        ('inp_type',       ('Signal', 'inp_type'),                   False, 'inp_type'),
        ('img_file',       ('Signal', 'File'),                       True,  'img_file'),
        ('img_dir',        ('Signal', 'img_dir'),                    True,  'img_dir'),
        ('include_subdir', ('Signal', 'include_subdir'),             False, 'include_subdir'),
        ('img_ext',        ('Signal', 'img_ext'),                    False, 'img_ext'),
        ('series_average', ('Signal', 'series_average'),             False, 'series_average'),
        ('meta_ext',       ('Signal', 'meta_ext'),                   False, None),
        ('file_filter',    ('Signal', 'Filter'),                     False, 'file_filter'),
        ('write_mode',     ('Signal', 'write_mode'),                 False, 'write_mode'),
        ('mask_file',      ('Signal', 'mask_file'),                  True,  'mask_file'),
        ('bg_type',        ('BG', 'bg_type'),                        False, 'bg_type'),
        ('bg_file',        ('BG', 'File'),                           True,  'bg_file'),
        ('bg_dir',         ('BG', 'Match', 'bg_dir'),                True,  'bg_dir'),
        ('bg_match_fname', ('BG', 'Match', 'match_fname'),           False, 'bg_match_fname'),
        ('bg_file_filter', ('BG', 'Match', 'Filter'),                False, 'bg_file_filter'),
        ('bg_scale',       ('BG', 'Scale'),                          False, 'bg_scale'),
        ('gi',             ('GI', 'Grazing'),                        False, 'gi'),
        ('th_mtr',         ('GI', 'th_motor'),                       False, 'th_mtr'),
        ('sample_orientation', ('GI', 'sample_orientation'),         False, 'sample_orientation'),
        ('tilt_angle',     ('GI', 'tilt_angle'),                     False, 'tilt_angle'),
        ('gi_mode_1d',     ('GI', 'gi_mode_1d'),                    False, 'gi_mode_1d'),
        ('gi_mode_2d',     ('GI', 'gi_mode_2d'),                    False, 'gi_mode_2d'),
        ('apply_threshold', ('Mask', 'Threshold'),                   False, 'apply_threshold'),
        ('threshold_min',  ('Mask', 'min'),                          False, 'threshold_min'),
        ('threshold_max',  ('Mask', 'max'),                          False, 'threshold_max'),
        ('h5_dir',         ('h5_dir',),                              True,  'h5_dir'),
    ]

    def _save_to_session(self, *args):
        data = {}
        for key, path, _, _ in self._SESSION_PARAMS:
            try:
                p = self.parameters
                for segment in path:
                    p = p.child(segment)
                data[key] = p.value()
            except Exception:
                pass
        data['processing_mode'] = self.ui.processingModeCombo.currentText()
        data['live_mode'] = self.ui.liveCheckBox.isChecked()
        data['batch_mode'] = self.ui.batchCheckBox.isChecked()
        save_session(data)

    def _restore_from_session(self):
        session = load_session()
        for key, path, is_path, attr in self._SESSION_PARAMS:
            val = session.get(key)
            if val is None:
                continue
            if is_path and not Path(val).exists():
                continue
            try:
                p = self.parameters
                for segment in path:
                    p = p.child(segment)
                p.setValue(val)
                if attr is not None:
                    setattr(self, attr, val)
            except Exception:
                pass
        # Restore processing mode dropdown and checkboxes
        mode = session.get('processing_mode')
        if mode:
            idx = self.ui.processingModeCombo.findText(mode)
            if idx >= 0:
                self.ui.processingModeCombo.setCurrentIndex(idx)
        if 'live_mode' in session:
            self.ui.liveCheckBox.setChecked(session['live_mode'])
        if 'batch_mode' in session:
            self.ui.batchCheckBox.setChecked(session['batch_mode'])
        # meta_ext needs None conversion (sigValueChanged fires set_meta_ext automatically)
        # poni_file needs poni_dict loaded
        if session.get('poni_file') and Path(session['poni_file']).exists():
            self.get_poni_dict()

    # Signal to notify static_scan_widget that viewer mode changed.
    # Emits the viewer_mode string ('image', 'xye') or '' for normal.
    sigViewerModeChanged = QtCore.Signal(str)

    def _on_mode_changed(self, *args):
        """Update all flags from the processing mode dropdown and checkboxes."""
        mode_text = self.ui.processingModeCombo.currentText()
        is_viewer = mode_text in ('Image Viewer', 'XYE Viewer')
        is_xye = mode_text == 'Int 1D (XYE)'

        # Pre-process state overrides
        self.ui.liveCheckBox.blockSignals(True)
        self.ui.batchCheckBox.blockSignals(True)

        if is_viewer:
            self.ui.liveCheckBox.setEnabled(False)
            self.ui.batchCheckBox.setEnabled(False)
            self.ui.coresLabel.setEnabled(False)
            self.ui.maxCoresSpinBox.setEnabled(False)
        elif is_xye:
            self.ui.liveCheckBox.setChecked(False)
            self.ui.liveCheckBox.setEnabled(False)
            self.ui.batchCheckBox.setChecked(True)
            self.ui.batchCheckBox.setEnabled(False)
            self.ui.coresLabel.setEnabled(True)
            self.ui.maxCoresSpinBox.setEnabled(True)
        else:
            self.ui.liveCheckBox.setEnabled(True)
            self.ui.batchCheckBox.setEnabled(True)

            # Mutual exclusion: when one is checked, uncheck the other
            is_live = self.ui.liveCheckBox.isChecked()
            is_batch = self.ui.batchCheckBox.isChecked()
            if is_live and is_batch:
                # Live was just checked — uncheck batch (or vice versa).
                # Determine which was the trigger by checking the sender.
                # If unclear, prefer live (last action wins).
                self.ui.batchCheckBox.setChecked(False)
                is_batch = False
            if is_live:
                self.ui.batchCheckBox.setEnabled(False)
            if is_batch:
                self.ui.liveCheckBox.setEnabled(False)

            # Sync cores enabled state with batch checkbox
            self.ui.coresLabel.setEnabled(is_batch)
            self.ui.maxCoresSpinBox.setEnabled(is_batch)

        self.ui.liveCheckBox.blockSignals(False)
        self.ui.batchCheckBox.blockSignals(False)

        # Ensure cores are always visible
        self.ui.coresLabel.setVisible(True)
        self.ui.maxCoresSpinBox.setVisible(True)

        self.batch_mode = self.ui.batchCheckBox.isChecked()
        self.live_mode = self.ui.liveCheckBox.isChecked()
        self.xye_only = is_xye
        
        if mode_text == 'Image Viewer':
            self.viewer_mode = 'image'
            self.sphere.skip_2d = False
        elif mode_text == 'XYE Viewer':
            self.viewer_mode = 'xye'
            self.sphere.skip_2d = False
        else:
            self.viewer_mode = None
            self.sphere.skip_2d = '1D' in mode_text

        # Sync to thread
        self.thread.batch_mode = self.batch_mode
        self.thread.xye_only = self.xye_only
        self.thread.live_mode = self.live_mode

        # Gray out integration controls in viewer mode
        self._set_integration_controls_enabled(not is_viewer)
        # Hide start/pause/stop/continue in viewer mode
        self.ui.frame.setVisible(not is_viewer)
        # Notify parent only when viewer mode actually changed (avoids
        # unnecessary layout resets when just toggling Live/Batch).
        new_vm = self.viewer_mode or ''
        if not hasattr(self, '_prev_viewer_mode'):
            self._prev_viewer_mode = ''
        if new_vm != self._prev_viewer_mode:
            self._prev_viewer_mode = new_vm
            self.sigViewerModeChanged.emit(new_vm)

    def _set_integration_controls_enabled(self, enabled):
        """Enable or disable parameter tree groups related to integration."""
        for group_name in ('Calibration', 'BG', 'Mask', 'GI'):
            try:
                grp = self.parameters.child(group_name)
                grp.setOpts(enabled=enabled)
            except Exception:
                pass
        # Also disable write mode and mask file in Signal group
        for child_name in ('write_mode', 'mask_file', 'mask_file_browse'):
            try:
                self.parameters.child('Signal').child(child_name).setOpts(enabled=enabled)
            except Exception:
                pass
        # Disable save path
        try:
            self.parameters.child('h5_dir').setOpts(enabled=enabled)
            self.parameters.child('h5_dir_browse').setOpts(enabled=enabled)
        except Exception:
            pass

    def setup(self):
        """Sets up the child thread, syncs all parameters.
        """
        # Calibration
        global ctr
        ctr += 1

        self.poni_file = self.parameters.child('Calibration').child('poni_file').value()
        self.thread.poni = self.poni

        # Signal
        self.file_filter = self.parameters.child('Signal').child('Filter').value()
        self.thread.file_filter = self.file_filter

        self.inp_type = self.parameters.child('Signal').child('inp_type').value()
        self.thread.inp_type = self.inp_type

        self.get_img_fname()
        self.thread.img_file = self.img_file

        self.scan_name, _ = _get_scan_info(self.img_file)
        self.thread.scan_name = self.scan_name

        self.thread.single_img = self.single_img
        self.thread.img_dir, self.thread.img_ext = self.img_dir, self.img_ext

        self.include_subdir = self.parameters.child('Signal').child('include_subdir').value()
        self.thread.include_subdir = self.include_subdir

        self.thread.series_average = self.series_average
        self.thread.meta_ext = self.meta_ext

        self.thread.h5_dir = self.h5_dir
        self.fname = os.path.join(self.h5_dir, self.scan_name + '.nxs')
        self.thread.fname = self.fname

        self.mask_file = self.parameters.child('Signal').child('mask_file').value()
        self.thread.mask_file = self.mask_file

        # Threshold
        self.apply_threshold = self.parameters.child('Mask').child('Threshold').value()
        self.thread.apply_threshold = self.apply_threshold
        self.threshold_min = self.parameters.child('Mask').child('min').value()
        self.thread.threshold_min = self.threshold_min
        self.threshold_max = self.parameters.child('Mask').child('max').value()
        self.thread.threshold_max = self.threshold_max

        # Write Mode
        self.write_mode = self.parameters.child('Signal').child('write_mode').value()
        self.thread.write_mode = self.write_mode

        # Background
        self.bg_type = self.parameters.child('BG').child('bg_type').value()
        self.thread.bg_type = self.bg_type

        self.bg_file = self.parameters.child('BG').child('File').value()
        self.thread.bg_file = self.bg_file

        self.bg_matching_par = self.parameters.child('BG').child('Match').child('Parameter').value()
        self.thread.bg_matching_par = self.bg_matching_par

        self.bg_dir = self.parameters.child('BG').child('Match').child('bg_dir').value()
        self.thread.bg_dir = self.bg_dir

        self.bg_match_fname = self.parameters.child('BG').child('Match').child('match_fname').value()
        self.thread.bg_match_fname = self.bg_match_fname

        self.bg_file_filter = self.parameters.child('BG').child('Match').child('Filter').value()
        self.thread.bg_file_filter = self.bg_file_filter

        self.bg_scale = self.parameters.child('BG').child('Scale').value()
        self.thread.bg_scale = self.bg_scale

        self.bg_norm_channel = self.parameters.child('BG').child('norm_channel').value()
        self.thread.bg_norm_channel = self.bg_norm_channel

        # Grazing Incidence
        self.gi = self.parameters.child('GI').child('Grazing').value()
        self.thread.gi = self.gi

        # self.th_mtr = self.parameters.child('GI').child('th_motor').value()
        self.thread.th_mtr = self.th_mtr

        self.sample_orientation = self.parameters.child('GI').child('sample_orientation').value()
        self.thread.sample_orientation = self.sample_orientation

        self.tilt_angle = self.parameters.child('GI').child('tilt_angle').value()
        self.thread.tilt_angle = self.tilt_angle

        # GI modes are driven by the integrator panel (axis1D / axis2D),
        # so read them from sphere.bai_*_args which the integrator updates.
        self.gi_mode_1d = self.sphere.bai_1d_args.get('gi_mode_1d', 'q_total')
        self.gi_mode_2d = self.sphere.bai_2d_args.get('gi_mode_2d', 'qip_qoop')
        self.thread.gi_mode_1d = self.gi_mode_1d
        self.thread.gi_mode_2d = self.gi_mode_2d

        # Notify integrator panel so labels/widgets update immediately.
        self.sigUpdateGI.emit(self.gi)

        # Processing mode flags and parallel cores
        self.thread.live_mode = self.live_mode
        self.thread.batch_mode = self.batch_mode
        self.thread.xye_only = self.xye_only
        self.thread.max_cores = self.ui.maxCoresSpinBox.value()
        self.sphere.max_cores = self.thread.max_cores  # used by sphere_threads

        self.thread.command = self.command

        self.thread.file_lock = self.file_lock
        self.thread.sphere_args = self.sphere_args

        self.thread.sphere = self.sphere
        self.thread.data_1d = self.data_1d
        self.thread.data_2d = self.data_2d

    def start(self):
        self.command = 'start'
        self.thread.command = 'start'
        self.ui.pauseButton.setEnabled(True)
        self.ui.continueButton.setEnabled(False)
        self.ui.stopButton.setEnabled(True)
        self.sigStart.emit()

    def pause(self):
        self.command = 'pause'
        self.thread.command = 'pause'
        # if self.thread.isRunning():
        self.ui.pauseButton.setEnabled(False)
        self.ui.continueButton.setEnabled(True)

    def cont(self):
        self.command = 'continue'
        self.thread.command = 'continue'
        # if self.thread.isRunning():
        self.ui.pauseButton.setEnabled(True)
        self.ui.continueButton.setEnabled(False)

    def stop(self):
        self.command = 'stop'
        self.thread.command = 'stop'
        # if self.thread.isRunning():
        self.ui.pauseButton.setEnabled(False)
        self.ui.continueButton.setEnabled(False)
        self.ui.stopButton.setEnabled(False)
        self.ui.specLabel.setText('')

    def set_poni_file(self):
        """Opens file dialogue and sets the calibration file
        """
        fname, _ = QFileDialog().getOpenFileName(
            filter="PONI (*.poni *.PONI)"
        )
        if fname != '':
            self.parameters.child('Calibration').child('poni_file').setValue(fname)
            self.poni_file = fname
            self._save_to_session()

    def get_poni_dict(self):
        """Load the PONI calibration file and store as a PONI object."""
        if not os.path.exists(self.poni_file):
            for child in self.parameters.children():
                child.hide()
            self.parameters.child('Calibration').show()
            return

        try:
            self.poni = PONI.from_poni_file(self.poni_file)
        except Exception:
            self.poni = None
        if self.poni is None:
            logger.warning('Invalid Poni File: %s', self.poni_file)
            self.thread.signal_q.put(('message', 'Invalid Poni File'))
            return

        for child in self.parameters.children():
            child.show()

    def set_inp_type(self):
        """Change Parameter Names depending on Input Type
        """
        self.single_img = False
        self.parameters.child('Signal').child('File').show()
        self.parameters.child('Signal').child('img_file_browse').show()
        self.parameters.child('Signal').child('img_dir').hide()
        self.parameters.child('Signal').child('img_dir_browse').hide()
        self.parameters.child('Signal').child('include_subdir').hide()
        self.parameters.child('Signal').child('Filter').hide()
        self.parameters.child('Signal').child('series_average').show()
        self.parameters.child('Signal').child('img_ext').hide()

        inp_type = self.parameters.child('Signal').child('inp_type').value()
        if inp_type == 'Image Directory':
            self.parameters.child('Signal').child('File').hide()
            self.parameters.child('Signal').child('img_file_browse').hide()
            self.parameters.child('Signal').child('img_dir').show()
            self.parameters.child('Signal').child('img_dir_browse').show()
            self.parameters.child('Signal').child('include_subdir').show()
            self.parameters.child('Signal').child('Filter').show()
            self.parameters.child('Signal').child('img_ext').show()

        if inp_type == 'Single Image':
            self.single_img = True
            self.parameters.child('Signal').child('series_average').hide()

        self.inp_type = inp_type
        self.get_img_fname()

    def set_img_file(self):
        """Opens file dialogue and sets the spec data file
        """
        fname, _ = QFileDialog().getOpenFileName(
            filter="Images (*.tiff *.tif *.h5 *.raw *.mar3450)"
        )
        if fname != '':
            self.parameters.child('Signal').child('File').setValue(fname)

    def set_img_dir(self):
        """Opens file dialogue and sets the signal data folder
        """
        path = QFileDialog().getExistingDirectory(
            caption='Choose Image Directory',
            dir='',
            options=QFileDialog.ShowDirsOnly
        )
        if path != '':
            self.parameters.child('Signal').child('img_dir').setValue(path)
        self.img_dir = path

    def get_img_fname(self):
        """Sets file name based on chosen options
        """
        old_fname = self.img_file
        if self.inp_type != 'Image Directory':
            img_file = self.parameters.child('Signal').child('File').value()
            if os.path.exists(img_file):
                self.img_file = img_file
                _p = Path(self.img_file)
                self.img_dir, self.img_ext = str(_p.parent), _p.suffix.lstrip('.')
                # Auto-detect metadata sidecar if not already set
                if not self.meta_ext:
                    self.detect_meta_ext(self.img_file)

        else:
            self.img_ext = self.parameters.child('Signal').child('img_ext').value()
            self.img_dir = self.parameters.child('Signal').child('img_dir').value()
            self.include_subdir = self.parameters.child('Signal').child('include_subdir').value()

            filters = '*' + '*'.join(f for f in self.file_filter.split()) + '*'
            filters = filters if filters != '**' else '*'

            file_found = False
            for idx, (subdir, dirs, files) in enumerate(os.walk(self.img_dir)):
                for file in files:
                    fname = os.path.join(subdir, file)
                    if fnmatch.fnmatch(fname, f'{filters}.{self.img_ext}'):
                        if match_img_detector(fname, self.poni):
                            if self.meta_ext:
                                if self.exists_meta_file(fname):
                                    self.img_file = fname
                                    file_found = True
                                    break
                                else:
                                    continue
                            else:
                                self.img_file = fname
                                # Auto-detect metadata sidecar if not set
                                if not self.meta_ext:
                                    self.detect_meta_ext(fname)
                                break
                if file_found or (not self.include_subdir):
                    break

        if (((self.img_file != old_fname) or (self.img_file and (len(self.scan_parameters) < 1)))
                and self.meta_ext):
            if self.exists_meta_file(self.img_file):
                self.set_pars_from_meta()

    def set_series_average(self):
        self.series_average = self.parameters.child('Signal').child('series_average').value()

    def set_meta_ext(self):
        self.meta_ext = self.parameters.child('Signal').child('meta_ext').value()
        if self.meta_ext == 'None':
            self.meta_ext = None
        self._save_to_session()
        self.get_img_fname()

    def exists_meta_file(self, img_file):
        """Checks for existence of meta file for image file"""
        if self.meta_ext != 'SPEC':
            meta_files = [
                f'{os.path.splitext(img_file)[0]}.{self.meta_ext}',
                f'{img_file}.{self.meta_ext}'
            ]
            if os.path.exists(meta_files[0]) or os.path.exists(meta_files[1]):
                return True
        else:
            spec_fname, _, _ = _extract_scan_info(Path(img_file))
            if spec_fname:
                img_fpath = Path(img_file)
                for parent in (img_fpath.parent, img_fpath.parents[1]):
                    if (parent / spec_fname).is_file():
                        return True

        return False

    def detect_meta_ext(self, img_file):
        """Auto-detect metadata sidecar format for *img_file*.

        Probes for ``.txt`` and ``.pdi`` sidecar files.  If found, updates
        the GUI parameter and ``self.meta_ext``.  Returns the detected
        extension string or ``None``.
        """
        base = os.path.splitext(img_file)[0]
        for ext in ('txt', 'pdi'):
            if os.path.exists(f'{base}.{ext}') or os.path.exists(f'{img_file}.{ext}'):
                # Update the GUI dropdown so the user sees the change
                param = self.parameters.child('Signal').child('meta_ext')
                param.setValue(ext)          # fires set_meta_ext automatically
                return ext
        return None

    def set_pars_from_meta(self):
        self.get_scan_parameters()
        self.set_bg_matching_options()
        self.set_gi_motor_options()
        self.set_bg_norm_options()

    def set_mask_file(self):
        """Opens file dialogue and sets the mask file
        """
        fname, _ = QFileDialog().getOpenFileName(
            filter="EDF (*.edf)"
        )
        if fname != '':
            self.parameters.child('Signal').child('mask_file').setValue(fname)
        self.mask_file = fname

    def set_bg_type(self):
        """Change Parameter Names depending on BG Type
        """
        for child in self.parameters.child('BG').children():
            child.hide()
        self.parameters.child('BG').child('bg_type').show()

        self.bg_type = self.parameters.child('BG').child('bg_type').value()
        if self.bg_type == 'None':
            return
        elif self.bg_type != 'BG Directory':
            self.parameters.child('BG').child('File').show()
            if self.bg_type == 'Single BG File':
                opts = {'title': 'File Name'}
            else:
                opts = {'title': 'First File'}
            self.parameters.child('BG').child('File').setOpts(**opts)
            self.parameters.child('BG').child('bg_file_browse').show()
        else:
            self.parameters.child('BG').child('Match').show()

        self.parameters.child('BG').child('Scale').show()
        self.parameters.child('BG').child('norm_channel').show()

    def set_bg_file(self):
        """Opens file dialogue and sets the background file
        """
        fname, _ = QFileDialog().getOpenFileName(
            filter=f"Images (*.{self.img_ext})"
        )
        if fname != '':
            self.parameters.child('BG').child('File').setValue(fname)
        self.bg_file = fname

    def set_bg_dir(self):
        """Opens file dialogue and sets the background folder
        """
        path = QFileDialog().getExistingDirectory(
            caption='Choose BG Directory',
            dir='',
            options=QFileDialog.ShowDirsOnly
        )
        if path != '':
            self.parameters.child('BG').child('Match').child('bg_dir').setValue(path)
        self.bg_dir = path

    def set_h5_dir(self):
        """Opens file dialogue and sets the path where processed data is stored
        """
        path = QFileDialog().getExistingDirectory(
            caption='Choose Save Directory',
            directory='',
            options=QFileDialog.ShowDirsOnly
        )
        if path != '':
            Path(path).mkdir(parents=True, exist_ok=True)
            self.parameters.child('h5_dir').setValue(path)
            self.h5_dir = path

    def set_bg_matching_options(self):
        """Reads image metadata to populate matching parameters
        """
        pars = [p for p in self.scan_parameters if not any(x.lower() in p.lower() for x in ['ROI', 'PD'])]
        pars.insert(0, 'None')
        if 'TEMP' in pars:
            pars.insert(1, pars.pop(pars.index('TEMP')))

        value = 'None'
        opts = {'values': pars, 'limits': pars, 'value': value}
        self.parameters.child('BG').child('Match').child('Parameter').setOpts(**opts)

    def set_bg_matching_par(self):
        """Changes bg matching parameter
        """
        self.bg_matching_par = self.parameters.child('BG').child('Match').child('Parameter').value()
        if self.bg_matching_par == 'None':
            self.bg_matching_par = None

    def set_bg_norm_options(self):
        """Counter Values used to normalize and subtract background
        """
        pars = self.counters
        pars.insert(0, 'None')

        opts = {'values': pars, 'limits': pars, 'value': 'None'}
        self.parameters.child('BG').child('norm_channel').setOpts(**opts)

    def set_bg_norm_channel(self):
        """Changes bg matching parameter
        """
        self.bg_norm_channel = self.parameters.child('BG').child('norm_channel').value()

    def set_gi_motor_options(self):
        """Reads image metadata to populate possible GI theta motor
        """
        pars = [p for p in self.motors if not any(x.lower() in p.lower() for x in ['ROI', 'PD'])]
        if 'th' in pars:
            pars.insert(0, pars.pop(pars.index('th')))
            value = 'th'
        elif 'theta' in pars:
            pars.insert(0, pars.pop(pars.index('theta')))
            value = 'theta'
        else:
            value = 'Theta'

        pars = ['Manual'] + pars

        opts = {'values': pars, 'limits': pars, 'value': value}
        self.parameters.child('GI').child('th_motor').setOpts(**opts)

    def set_gi_th_motor(self):
        """Update Grazing theta motor"""
        self.th_mtr = self.parameters.child('GI').child('th_motor').value()
        self.parameters.child('GI').child('th_val').hide()
        if self.th_mtr == 'Manual':
            self.parameters.child('GI').child('th_val').show()
            self.th_mtr = self.parameters.child('GI').child('th_val').value()

    def get_scan_parameters(self):
        """ Reads image metadata to populate matching parameters
        """
        if not self.img_file:
            return

        img_meta = read_image_metadata(self.img_file, meta_format=self.meta_ext)
        self.scan_parameters = list(img_meta.keys())
        self.counters = self.scan_parameters
        self.motors = self.scan_parameters

    def enabled(self, enable):
        """Sets tree and start button to enable.

        args:
            enable: bool, True for enabled False for disabled.
        """
        self.tree.setEnabled(enable)
        self.ui.startButton.setEnabled(enable)

    def stylize_ParameterTree(self):
        self.tree.setStyleSheet("""
        QTreeView::item:has-children {
            background-color: rgb(230, 230, 230); 
            color: rgb(30, 30, 30);
        }
            """)





# -*- coding: utf-8 -*-
"""
Display constants — axis labels, units, and helper functions shared
across display_frame_widget and related modules.

Extracted from display_frame_widget.py to keep configuration separate
from widget logic.

@author: thampy
"""

import numpy as np
from .integrator import GI_MODES_1D, GI_MODES_2D  # noqa: F401

# ── Unicode symbols ──
AA_inv = u'\u212B\u207B\u00B9'
Th = u'\u03B8'
Chi = u'\u03C7'
Deg = u'\u00B0'
Qip_s = u'Q\u1D62\u209A'        # Q_ip subscript
Qoop_s = u'Q\u2092\u2092\u209A'  # Q_oop subscript
Qtot_s = u'Q\u209C\u2092\u209C'  # Q_tot subscript

# ── Standard (non-GI) axis options ──
plotUnits = [f"Q ({AA_inv})", f"2{Th} ({Deg})", f"{Chi} ({Deg})"]
imageUnits = [f"Q-{Chi}", f"2{Th}-{Chi}"]

x_labels_1D = ('Q', f"2{Th}", Chi)
x_units_1D = (AA_inv, Deg, Deg)

x_labels_2D = ('Q', f"2{Th}")
x_units_2D = (AA_inv, Deg)

y_labels_2D = (Chi, Chi)
y_units_2D = (Deg, Deg)

# ── GI axis options (order matches GI_MODES_1D / GI_MODES_2D) ──
gi_plotUnits = [
    f"Q ({AA_inv})",
    f"{Qip_s} ({AA_inv})",
    f"{Qoop_s} ({AA_inv})",
    f"Exit angle ({Deg})",
]
gi_imageUnits = [
    f"{Qip_s}-{Qoop_s}",
    f"Q-{Chi}",
    f"Exit angles",
]

gi_x_labels_1D = ('Q', Qip_s, Qoop_s, 'Exit angle')
gi_x_units_1D = (AA_inv, AA_inv, AA_inv, Deg)

gi_x_labels_2D = (Qip_s, 'Q', 'Exit angle')
gi_x_units_2D = (AA_inv, AA_inv, Deg)

gi_y_labels_2D = (Qoop_s, Chi, 'Exit angle')
gi_y_units_2D = (AA_inv, Deg, Deg)

# ── Axis decomposition for 2D → 1D slicing ──
# Maps each GI 2D mode to the pair of 1D axis labels available from it.
# The keys are GI_MODES_2D entries; values are (radial_label, azimuthal_label)
# using the same gi_plotUnits label strings for consistency.
GI_2D_AXES = {
    'qip_qoop': (
        f"{Qip_s} ({AA_inv})",   # radial  (x axis of 2D)
        f"{Qoop_s} ({AA_inv})",  # azimuthal (y axis of 2D)
    ),
    'q_chi': (
        f"Q ({AA_inv})",         # radial
        f"{Chi} ({Deg})",        # azimuthal
    ),
    'exit_angles': (
        f"Exit angle ({Deg})",   # both axes are exit angles
        f"Exit angle ({Deg})",
    ),
}

# Standard (non-GI) 2D axes
STD_2D_AXES = {
    0: (f"Q ({AA_inv})", f"{Chi} ({Deg})"),       # Q-Chi image
    1: (f"2{Th} ({Deg})", f"{Chi} ({Deg})"),       # 2Th-Chi image
}


def _downsample_for_display(data, widget):
    """Reduce array resolution to match widget pixel size. Display-only."""
    if data is None or data.ndim != 2:
        return data
    w = max(widget.width(), 200)
    h = max(widget.height(), 200)
    if data.shape[0] > w * 2 or data.shape[1] > h * 2:
        sx = max(1, data.shape[0] // w)
        sy = max(1, data.shape[1] // h)
        return data[::sx, ::sy]
    return data



"""
xdart container types — thin re-export from ssrl_xrd_tools.
"""
from ssrl_xrd_tools.core.containers import (
    PONI,
    IntegrationResult1D,
    IntegrationResult2D,
)

__all__ = [
    "PONI",
    "IntegrationResult1D",
    "IntegrationResult2D",
]



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'staticUI.ui'
##
## Created by: Qt User Interface Compiler version 6.10.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QFrame, QHBoxLayout,
    QSizePolicy, QSplitter, QStackedWidget, QVBoxLayout,
    QWidget)

class Ui_Form(object):
    def setupUi(self, Form):
        if not Form.objectName():
            Form.setObjectName(u"Form")
        Form.resize(1440, 842)
        self.horizontalLayout = QHBoxLayout(Form)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalLayout.setContentsMargins(0, 0, 0, 0)
        self.mainSplitter = QSplitter(Form)
        self.mainSplitter.setObjectName(u"mainSplitter")
        self.mainSplitter.setOrientation(Qt.Horizontal)
        self.leftFrame = QFrame(self.mainSplitter)
        self.leftFrame.setObjectName(u"leftFrame")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(1)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.leftFrame.sizePolicy().hasHeightForWidth())
        self.leftFrame.setSizePolicy(sizePolicy)
        self.leftFrame.setMinimumSize(QSize(200, 0))
        self.leftFrame.setFrameShape(QFrame.StyledPanel)
        self.leftFrame.setFrameShadow(QFrame.Raised)
        self.leftFrame.setLineWidth(5)
        self.horizontalLayout_4 = QHBoxLayout(self.leftFrame)
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.horizontalLayout_4.setContentsMargins(0, 0, 0, 0)
        self.leftSplitter = QSplitter(self.leftFrame)
        self.leftSplitter.setObjectName(u"leftSplitter")
        self.leftSplitter.setOrientation(Qt.Vertical)
        self.hdf5Frame = QFrame(self.leftSplitter)
        self.hdf5Frame.setObjectName(u"hdf5Frame")
        self.hdf5Frame.setMinimumSize(QSize(0, 0))
        self.hdf5Frame.setFrameShape(QFrame.StyledPanel)
        self.hdf5Frame.setFrameShadow(QFrame.Raised)
        self.hdf5Frame.setLineWidth(3)
        self.leftSplitter.addWidget(self.hdf5Frame)
        self.metaFrame = QFrame(self.leftSplitter)
        self.metaFrame.setObjectName(u"metaFrame")
        self.metaFrame.setFrameShape(QFrame.StyledPanel)
        self.metaFrame.setFrameShadow(QFrame.Raised)
        self.metaFrame.setLineWidth(3)
        self.leftSplitter.addWidget(self.metaFrame)

        self.horizontalLayout_4.addWidget(self.leftSplitter)

        self.mainSplitter.addWidget(self.leftFrame)
        self.middleFrame = QFrame(self.mainSplitter)
        self.middleFrame.setObjectName(u"middleFrame")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy1.setHorizontalStretch(2)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.middleFrame.sizePolicy().hasHeightForWidth())
        self.middleFrame.setSizePolicy(sizePolicy1)
        self.middleFrame.setFrameShape(QFrame.StyledPanel)
        self.middleFrame.setFrameShadow(QFrame.Raised)
        self.middleFrame.setLineWidth(5)
        self.mainSplitter.addWidget(self.middleFrame)
        self.rightFrame = QFrame(self.mainSplitter)
        self.rightFrame.setObjectName(u"rightFrame")
        sizePolicy.setHeightForWidth(self.rightFrame.sizePolicy().hasHeightForWidth())
        self.rightFrame.setSizePolicy(sizePolicy)
        self.rightFrame.setFrameShape(QFrame.StyledPanel)
        self.rightFrame.setFrameShadow(QFrame.Raised)
        self.rightFrame.setLineWidth(5)
        self.horizontalLayout_3 = QHBoxLayout(self.rightFrame)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.horizontalLayout_3.setContentsMargins(0, 0, 0, 0)
        self.rightSplitter = QSplitter(self.rightFrame)
        self.rightSplitter.setObjectName(u"rightSplitter")
        self.rightSplitter.setOrientation(Qt.Vertical)
        self.integratorFrame = QFrame(self.rightSplitter)
        self.integratorFrame.setObjectName(u"integratorFrame")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.integratorFrame.sizePolicy().hasHeightForWidth())
        self.integratorFrame.setSizePolicy(sizePolicy2)
        self.integratorFrame.setMaximumSize(QSize(16777215, 300))
        self.integratorFrame.setFrameShape(QFrame.StyledPanel)
        self.integratorFrame.setFrameShadow(QFrame.Raised)
        self.integratorFrame.setLineWidth(3)
        self.rightSplitter.addWidget(self.integratorFrame)
        self.wranglerFrame = QFrame(self.rightSplitter)
        self.wranglerFrame.setObjectName(u"wranglerFrame")
        sizePolicy3 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Expanding)
        sizePolicy3.setHorizontalStretch(0)
        sizePolicy3.setVerticalStretch(0)
        sizePolicy3.setHeightForWidth(self.wranglerFrame.sizePolicy().hasHeightForWidth())
        self.wranglerFrame.setSizePolicy(sizePolicy3)
        self.wranglerFrame.setFrameShape(QFrame.StyledPanel)
        self.wranglerFrame.setFrameShadow(QFrame.Raised)
        self.wranglerFrame.setLineWidth(3)
        self.verticalLayout = QVBoxLayout(self.wranglerFrame)
        self.verticalLayout.setSpacing(0)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.wranglerStack = QStackedWidget(self.wranglerFrame)
        self.wranglerStack.setObjectName(u"wranglerStack")
        sizePolicy3.setHeightForWidth(self.wranglerStack.sizePolicy().hasHeightForWidth())
        self.wranglerStack.setSizePolicy(sizePolicy3)

        self.verticalLayout.addWidget(self.wranglerStack)

        self.rightSplitter.addWidget(self.wranglerFrame)

        self.horizontalLayout_3.addWidget(self.rightSplitter)

        self.mainSplitter.addWidget(self.rightFrame)

        self.horizontalLayout.addWidget(self.mainSplitter)


        self.retranslateUi(Form)

        QMetaObject.connectSlotsByName(Form)
    # setupUi

    def retranslateUi(self, Form):
        Form.setWindowTitle(QCoreApplication.translate("Form", u"Form", None))
    # retranslateUi




# -*- coding: utf-8 -*-
"""
Created on Mon Aug 26 14:21:58 2019

@author: walroth
"""
import copy
from threading import Condition

from pyFAI.integrator.azimuthal import AzimuthalIntegrator
from pyFAI import units
import numpy as np

from scipy.ndimage import zoom as ndimage_zoom

from xdart import utils
from ssrl_xrd_tools.core.containers import (
    PONI,
    IntegrationResult1D,
    IntegrationResult2D,
)
from ssrl_xrd_tools.integrate.calibration import poni_to_integrator
from ssrl_xrd_tools.integrate.single import integrate_1d, integrate_2d
from ssrl_xrd_tools.integrate.gid import (
    create_fiber_integrator,
    integrate_gi_1d,
    integrate_gi_2d,
    integrate_gi_polar,
    integrate_gi_polar_1d,
    integrate_gi_exitangles_1d,
    integrate_gi_exitangles,
)



# Maximum thumbnail dimension — both axes are scaled to fit within this.
_THUMBNAIL_MAX = 256


def _make_thumbnail(image, mask_idx=None, global_mask=None, max_size=_THUMBNAIL_MAX):
    """Downsample a 2D image to at most (max_size, max_size) using scipy zoom.

    Masked pixels are set to NaN *before* downsampling so the mask is
    baked into the thumbnail.  This lets the GUI display a masked preview
    without needing the full-resolution mask array.

    Parameters
    ----------
    image : ndarray
        2D detector image (typically map_raw - bg_raw).
    mask_idx : array-like or None
        Flat indices of pixels to mask (arch-level mask).
    global_mask : array-like or None
        Flat indices of pixels to mask (detector-level mask).
    max_size : int
        Maximum dimension of the output thumbnail.

    Returns float32 array, or None if input is invalid.
    """
    if image is None:
        return None
    arr = np.asarray(image, dtype=np.float32)
    if arr.ndim != 2:
        return None

    # Apply masks as NaN before downsampling
    all_mask = []
    if mask_idx is not None and len(mask_idx) > 0:
        all_mask.append(np.asarray(mask_idx, dtype=int))
    if global_mask is not None and len(global_mask) > 0:
        all_mask.append(np.asarray(global_mask, dtype=int))
    if all_mask:
        flat_mask = np.unique(np.concatenate(all_mask))
        flat_mask = flat_mask[flat_mask < arr.size]  # safety bound
        arr_flat = arr.ravel()
        arr_flat[flat_mask] = np.nan
        arr = arr_flat.reshape(arr.shape)

    h, w = arr.shape
    if h <= max_size and w <= max_size:
        return arr
    factor = min(max_size / h, max_size / w)
    return ndimage_zoom(arr, factor, order=1).astype(np.float32)


class EwaldArch():
    """Class for storing area detector data collected in
    X-ray diffraction experiments.

    Attributes:
        ai_args: dict, arguments passed to AzimuthalIntegrator
        arch_lock: Condition, threading lock used to ensure only one
            process can access data at a time
        file_lock: Condition, lock to ensure only one writer to
            data file
        idx: int, integer name of arch
        int_1d: int_1d_data/_static object from containers (for scanning/static detectors)
        int_2d: int_2d_data/_static object from containers (for scanning/static detectors)
        integrator: AzimuthalIntegrator object from pyFAI
        map_raw: numpy 2d array of the unprocessed image data
        bg_raw: numpy 2d array of the unprocessed image data for BG
        map_norm: float, normalization constant
        mask: numpy array of indeces to be masked in array.
        poni: poni data for integration
        scan_info: dict, information from any relevant motors and
            sensors
        static: bool, flag to specify if detector is static
        gi: bool, flag to specify if scattering geometry is grazing incidence
        th_mtr: str, the motor that controls sample rotation in gi mode
        tilt_angle: float, chi offset in gi geometry
        series_average: bool, flag to specify if series of images is averaged

    Methods:
        copy: create copy of arch
        get_mask: return mask array to feed into integrate1d
        integrate_1d: integrate the image data, results stored in
            int_1d_data
        integrate_2d: integrate the image data, results stored in
            int_2d_data
        load_from_h5: load data from hdf5 file
        save_to_h5: save data to hdf5 file
        set_integrator: set new integrator
        set_map_raw: replace raw data
        set_mask: replace mask data
        set_poni: replace poni object
        set_scan_info: replace scan_info
    """
    # pylint: disable=too-many-instance-attributes

    def __init__(self, idx=None, map_raw=None, poni=None, mask=None,
                 scan_info={}, ai_args={}, file_lock=Condition(),
                 static=False, bg_raw=0,
                 gi=False, th_mtr='th', tilt_angle=0,
                 sample_orientation=4,
                 series_average=False, integrator=None
                 ):
        # pylint: disable=too-many-arguments
        """idx: int, name of the arch.
        map_raw: numpy array, raw image data
        bg_raw: numpy array, raw image data for BG
        poni: PONI object, calibration data
        mask: None or numpy array, indices of pixels to mask
        scan_info: dict, metadata about scan
        ai_args: dict, args to be fed to azimuthalIntegrator constructor
        file_lock: Condition, lock for file access.
        """
        super(EwaldArch, self).__init__()
        self.idx = idx
        self.map_raw = map_raw
        self.bg_raw = bg_raw
        if poni is None:
            self.poni = PONI(dist=0.0, poni1=0.0, poni2=0.0)
        else:
            self.poni = poni
        if mask is None and map_raw is not None:
            self.mask = np.arange(map_raw.size)[map_raw.flatten() < 0]
        else:
            self.mask = mask
        self.scan_info = scan_info
        self.ai_args = ai_args
        self.file_lock = file_lock

        self.static = static
        self.gi = gi
        self.th_mtr = th_mtr
        self.tilt_angle = tilt_angle
        self.sample_orientation = sample_orientation
        self.series_average = series_average

        self.integrator = integrator if integrator is not None else self.setup_integrator()
        # Disable pyFAI's legacy mask heuristic that auto-inverts masks
        # with > 50% masked pixels (e.g. threshold masks).
        self.integrator.USE_LEGACY_MASK_NORMALIZATION = False

        self.arch_lock = Condition()
        self.map_norm = 1

        self.int_1d: IntegrationResult1D | None = None
        self.int_2d: IntegrationResult2D | None = None
        self.gi_1d: dict[str, IntegrationResult1D] = {}
        self.gi_2d: dict[str, IntegrationResult2D] = {}
        self.source_file: str = ""
        self.thumbnail: np.ndarray | None = None  # downsampled (map_raw - bg_raw)

    def __getstate__(self):
        """Exclude threading.Condition objects so EwaldArch can be pickled
        for use with concurrent.futures.ProcessPoolExecutor."""
        state = self.__dict__.copy()
        state.pop('arch_lock', None)
        state.pop('file_lock', None)
        return state

    def __setstate__(self, state):
        """Restore threading.Condition objects after unpickling."""
        self.__dict__.update(state)
        self.arch_lock = Condition()
        self.file_lock = Condition()

    def setup_integrator(self):
        """Sets up integrator object (always a plain AzimuthalIntegrator;
        GI uses create_fiber_integrator transiently during integration)."""
        return poni_to_integrator(self.poni)

    def _poni_from_integrator(self):
        """Return an ssrl_xrd_tools PONI derived from self.integrator's current geometry."""
        ai = self.integrator
        det = getattr(ai, 'detector', None)
        det_name = getattr(det, 'name', '') or ''
        wl = getattr(ai, 'wavelength', None)
        return PONI(
            dist=float(ai.dist),
            poni1=float(ai.poni1),
            poni2=float(ai.poni2),
            rot1=float(ai.rot1),
            rot2=float(ai.rot2),
            rot3=float(ai.rot3),
            wavelength=float(wl) if wl else 0.0,
            detector=str(det_name),
        )

    def _get_incident_angle(self):
        """Return incident angle in degrees from th_mtr or scan_info."""
        try:
            return float(self.th_mtr)
        except (ValueError, TypeError):
            # Case-insensitive lookup for motor position
            th_mtr = str(self.th_mtr).lower()
            for key, val in self.scan_info.items():
                if key.lower() == th_mtr:
                    return float(val)
            return 0.0

    def reset(self):
        """Clears all data, resets to a default EwaldArch.
        """
        self.idx = None
        self.map_raw = None
        self.bg_raw = None
        self.poni = PONI(dist=0.0, poni1=0.0, poni2=0.0)
        self.mask = None
        self.scan_info = {}
        self.integrator = self.setup_integrator()
        self.map_norm = 1
        self.int_1d = None
        self.int_2d = None
        self.gi_1d = {}
        self.gi_2d = {}
            
    def get_mask(self, global_mask=None):
        shape = self.map_raw.shape
        size = self.map_raw.size
        mask = np.zeros(size, dtype=bool)
        try:
            if self.mask is not None and len(self.mask):
                mask[self.mask] = True
            if global_mask is not None and len(global_mask):
                mask[global_mask] = True
            return mask.reshape(shape)
        except IndexError:
            print('Mask File Shape Mismatch')
            return mask.reshape(shape) if mask.size == size else mask

    def integrate_1d(self, numpoints=10000, radial_range=None,
                     monitor=None, unit=units.TTH_DEG, global_mask=None,
                     fiber_integrator=None, **kwargs):
        """Wrapper for integrate1d method of AzimuthalIntegrator from pyFAI.
        Returns result and also stores the data in the int_1d object.

        args:
            numpoints: int, number of points in final array
            radial_range: tuple or list, lower and upper end of integration
            monitor: str, keyword for normalization counter in scan_info
            unit: pyFAI unit for integration, units.TTH_DEG, units.Q_A,
                '2th_deg', or 'q_A^-1'
            kwargs: other keywords to be passed to integrate1d, see pyFAI docs.

        returns:
            result: integrate1d result from pyFAI.
        """
        with self.arch_lock:
            if self.map_raw is None:
                return  # no raw data (e.g. loaded from NeXus) — cannot re-integrate

            self.map_norm = 1
            if monitor is not None:
                if monitor.upper() in self.scan_info.keys():
                    self.map_norm = self.scan_info[monitor.upper()]
                elif monitor.lower() in self.scan_info.keys():
                    self.map_norm = self.scan_info[monitor.lower()]

            if self.mask is None:
                self.mask = np.arange(self.map_raw.size)[self.map_raw.flatten() < 0]

            # Keys that only make sense for GI integration — strip before
            # passing to the standard pyFAI integrator.
            _gi_only_keys = {
                'gi_mode_1d', 'gi_mode_2d', 'npt_oop',
                'sample_orientation', 'tilt_angle',
            }

            if not self.gi:
                std_kwargs = {k: v for k, v in kwargs.items()
                              if k not in _gi_only_keys}
                chi_offset = std_kwargs.pop('chi_offset', 0.0)
                if 'azimuth_range' in std_kwargs and std_kwargs['azimuth_range'] is not None:
                    _az = std_kwargs['azimuth_range']
                    std_kwargs['azimuth_range'] = (_az[0] - chi_offset, _az[1] - chi_offset)
                
                result = integrate_1d(
                    (self.map_raw - self.bg_raw) / self.map_norm,
                    self.integrator,
                    npt=numpoints,
                    unit=str(unit),
                    radial_range=radial_range,
                    mask=self.get_mask(global_mask),
                    **std_kwargs,
                )
                self.int_1d = result
            else:
                _gi_valid = {
                    'correctSolidAngle', 'variance', 'error_model',
                    'dummy', 'delta_dummy', 'polarization_factor', 'dark', 'flat',
                    'normalization_factor',
                }
                gi_kwargs = {k: v for k, v in kwargs.items() if k in _gi_valid}

                incident_angle = self._get_incident_angle()
                fi = fiber_integrator or create_fiber_integrator(
                    self._poni_from_integrator(),
                    incident_angle=incident_angle,
                    tilt_angle=self.tilt_angle,
                    sample_orientation=self.sample_orientation,
                    angle_unit="deg",
                )
                self._fiber_integrator = fi  # cache for integrate_2d

                image_data = (self.map_raw - self.bg_raw) / self.map_norm
                mask = self.get_mask(global_mask)
                gi_mode_1d = kwargs.get('gi_mode_1d', 'q_total')
                npt_oop = kwargs.get('npt_oop', numpoints)

                # Only compute the selected GI 1D mode
                if gi_mode_1d == 'q_ip':
                    # IP profile: integrate over OOP, return Q_ip
                    result = integrate_gi_1d(
                        image_data, fi, npt=numpoints, npt_oop=npt_oop,
                        unit='qip_A^-1',
                        method='no', mask=mask,
                        radial_range=radial_range,
                        azimuth_range=kwargs.get('azimuth_range'),
                        vertical_integration=False,
                        **gi_kwargs,
                    )
                    self.gi_1d['qip'] = result
                elif gi_mode_1d == 'q_oop':
                    # OOP profile: integrate over IP, return Q_oop
                    result = integrate_gi_1d(
                        image_data, fi, npt=numpoints, npt_oop=npt_oop,
                        unit='qoop_A^-1',
                        method='no', mask=mask,
                        radial_range=radial_range,
                        azimuth_range=kwargs.get('azimuth_range'),
                        vertical_integration=True,
                        **gi_kwargs,
                    )
                    self.gi_1d['qoop'] = result
                elif gi_mode_1d == 'exit_angle':
                    result = integrate_gi_exitangles_1d(
                        image_data, fi, npt=numpoints,
                        method='no', mask=mask, 
                        radial_range=radial_range,
                        azimuth_range=kwargs.get('azimuth_range'),
                        **gi_kwargs,
                    )
                    self.gi_1d['exit'] = result
                else:  # 'q_total' (default — polar integration)
                    result = integrate_gi_polar_1d(
                        image_data, fi, npt=numpoints,
                        method='no', mask=mask,
                        radial_range=radial_range,
                        azimuth_range=kwargs.get('azimuth_range'),
                        **gi_kwargs,
                    )
                    self.gi_1d['qtotal'] = result

                self.int_1d = result

    def integrate_2d(self, npt_rad=1000, npt_azim=1000, monitor=None,
                     radial_range=None, azimuth_range=None,
                     x_range=None, y_range=None,
                     unit=units.TTH_DEG, global_mask=None,
                     fiber_integrator=None, **kwargs):
        """Wrapper for integrate2d method of AzimuthalIntegrator from pyFAI.
        Returns result and also stores the data in the int_2d object.

        args:
            npt_rad: int, number of points in radial dimension. If
                None, will take number from the shape of map_norm
            npt_azim: int, number of points in azimuthal dimension. If
                None, will take number from the shape of map_norm
            radial_range: tuple or list, lower and upper end of integration
            azimuth_range: tuple or list, lower and upper end of integration
                in azimuthal direction
            monitor: str, keyword for normalization counter in scan_info
            unit: pyFAI unit for integration, units.TTH_DEG, units.Q_A,
                '2th_deg', or 'q_A^-1'
            x_range: tuple or list, IP range for GI reciprocal map
            y_range: tuple or list, OOP range for GI reciprocal map
            kwargs: other keywords to be passed to integrate2d, see pyFAI docs.

        returns:
            result: integrate2d result from pyFAI.
        """
        with self.arch_lock:
            if self.map_raw is None:
                return  # no raw data (e.g. loaded from NeXus) — cannot re-integrate

            if monitor is not None:
                self.map_norm = 1

            if self.mask is None:
                self.mask = np.arange(self.map_raw.size)[self.map_raw.flatten() < 0]

            if npt_rad is None:
                npt_rad = self.map_raw.shape[0]

            if npt_azim is None:
                npt_azim = self.map_raw.shape[1]

            _gi_only_keys = {
                'gi_mode_1d', 'gi_mode_2d', 'npt_oop',
                'sample_orientation', 'tilt_angle',
            }

            if not self.gi:
                std_kwargs = {k: v for k, v in kwargs.items()
                              if k not in _gi_only_keys}
                chi_offset = std_kwargs.pop('chi_offset', 0.0)
                if azimuth_range is not None:
                    azimuth_range = (azimuth_range[0] - chi_offset, azimuth_range[1] - chi_offset)
                
                result = integrate_2d(
                    (self.map_raw - self.bg_raw) / self.map_norm,
                    self.integrator,
                    npt_rad=npt_rad,
                    npt_azim=npt_azim,
                    unit=str(unit),
                    mask=self.get_mask(global_mask),
                    radial_range=radial_range,
                    azimuth_range=azimuth_range,
                    **std_kwargs,
                )
                
                # Apply explicit chi_offset rather than guessing midpoint
                if len(result.azimuthal) > 0:
                    result.azimuthal = result.azimuthal + chi_offset
                self.int_2d = result
            else:
                _gi_valid = {
                    'correctSolidAngle', 'variance', 'error_model',
                    'dummy', 'delta_dummy', 'polarization_factor', 'dark', 'flat',
                    'normalization_factor',
                }
                gi_kwargs = {k: v for k, v in kwargs.items() if k in _gi_valid}

                # Re-use FiberIntegrator cached by integrate_1d when available
                fi = fiber_integrator or getattr(self, '_fiber_integrator', None)
                if fi is None:
                    incident_angle = self._get_incident_angle()
                    fi = create_fiber_integrator(
                        self._poni_from_integrator(),
                        incident_angle=incident_angle,
                        tilt_angle=self.tilt_angle,
                        sample_orientation=self.sample_orientation,
                        angle_unit="deg",
                    )

                image_data = (self.map_raw - self.bg_raw) / self.map_norm
                mask = self.get_mask(global_mask)
                gi_mode_2d = kwargs.get('gi_mode_2d', 'qip_qoop')

                # Only compute the selected GI 2D mode
                if gi_mode_2d == 'q_chi':
                    result = integrate_gi_polar(
                        image_data, fi, npt_rad=npt_rad, npt_azim=npt_azim,
                        method='no', mask=mask,
                        radial_range=radial_range,
                        azimuth_range=azimuth_range,
                        **gi_kwargs,
                    )
                    self.gi_2d['polar'] = result
                elif gi_mode_2d == 'exit_angles':
                    result = integrate_gi_exitangles(
                        image_data, fi, npt_rad=npt_rad, npt_azim=npt_azim,
                        method='no', mask=mask,
                        radial_range=radial_range,
                        azimuth_range=azimuth_range,
                        **gi_kwargs,
                    )
                    self.gi_2d['exit2d'] = result
                else:  # 'qip_qoop' (default)
                    r_gi2d = integrate_gi_2d(
                        image_data, fi, npt_rad=npt_rad, npt_azim=npt_azim,
                        method='no', mask=mask,
                        radial_range=x_range, azimuth_range=y_range,
                        **gi_kwargs,
                    )
                    # integrate_gi_2d returns (npt_ip, npt_oop) via
                    # _to_result_2d transpose.  Use directly — the display
                    # code skips the azimuthal flip for qip_qoop mode.
                    result = r_gi2d
                    self.gi_2d['gi2d'] = result

                self.int_2d = result

    def set_integrator(self, **args):
        """Sets AzimuthalIntegrator with new arguments.

        args:
            args: see pyFAI for acceptable arguments for the integrator
                constructor.

        returns:
            None
        """

        with self.arch_lock:
            self.ai_args = args
            self.integrator = self.setup_integrator()

    def set_map_raw(self, new_data):
        with self.arch_lock:
            self.map_raw = new_data
            if self.mask is None:
                self.mask = np.arange(new_data.size)[new_data.flatten() < 0]

    def set_poni(self, new_data):
        with self.arch_lock:
            self.poni = new_data

    def set_mask(self, new_data):
        with self.arch_lock:
            self.mask = new_data

    def set_scan_info(self, new_data):
        with self.arch_lock:
            self.scan_info = new_data

    def save_to_h5(self, file, compression=None):
        """Saves data to hdf5 file using h5py as backend.
        DEPRECATED — use save_to_nexus for new code.

        args:
            file: h5py group or file object.

        returns:
            None
        """
        if self.static:
            compression = None
        with self.file_lock:
            if str(self.idx) in file:
                grp = file[str(self.idx)]
            else:
                grp = file.create_group(str(self.idx))
            grp.attrs['type'] = 'EwaldArch'
            if getattr(self, 'skip_map_raw', False):
                lst_attr = [
                    "mask", "map_norm", "scan_info", "ai_args",
                    "gi", "static"
                ]
            else:
                lst_attr = [
                    "map_raw", "mask", "map_norm", "scan_info", "ai_args",
                    "gi", "static", "bg_raw"
                ]
            utils.attributes_to_h5(self, grp, lst_attr,
                                   compression=compression)
            if self.int_1d is not None:
                if 'int_1d' not in grp:
                    grp.create_group('int_1d')
                self.int_1d.to_hdf5(grp['int_1d'], compression)
            if self.int_2d is not None:
                if 'int_2d' not in grp:
                    grp.create_group('int_2d')
                self.int_2d.to_hdf5(grp['int_2d'], compression)
            # Save auxiliary GI results
            if self.gi_1d:
                gi1d_grp = grp.require_group('gi_1d')
                for key, res in self.gi_1d.items():
                    sub = gi1d_grp.require_group(key)
                    res.to_hdf5(sub, compression)
            if self.gi_2d:
                gi2d_grp = grp.require_group('gi_2d')
                for key, res in self.gi_2d.items():
                    sub = gi2d_grp.require_group(key)
                    res.to_hdf5(sub, compression)
            if 'poni' not in grp:
                grp.create_group('poni')
            utils.dict_to_h5(self.poni.to_dict(), grp, 'poni')

    def load_from_h5(self, file, load_2d=True):
        """Loads data from hdf5 file and sets attributes.
        DEPRECATED — use load_from_nexus for new code.

        args:
            file: h5py file or group object
        """
        with self.file_lock:
            with self.arch_lock:
                if str(self.idx) not in file:
                    print("No data can be found")
                else:
                    grp = file[str(self.idx)]
                    if 'type' in grp.attrs:
                        if grp.attrs['type'] == 'EwaldArch':
                            lst_attr = [
                                "map_raw", "mask", "map_norm", "scan_info", "ai_args",
                                "gi", "static", "bg_raw"
                            ]
                            utils.h5_to_attributes(self, grp, lst_attr)
                            if 'int_1d' in grp:
                                self.int_1d = IntegrationResult1D.from_hdf5(grp['int_1d'])
                            if load_2d and 'int_2d' in grp:
                                self.int_2d = IntegrationResult2D.from_hdf5(grp['int_2d'])
                            # Load auxiliary GI results if present
                            if 'gi_1d' in grp:
                                for key in grp['gi_1d']:
                                    self.gi_1d[key] = IntegrationResult1D.from_hdf5(
                                        grp['gi_1d'][key]
                                    )
                            if load_2d and 'gi_2d' in grp:
                                for key in grp['gi_2d']:
                                    self.gi_2d[key] = IntegrationResult2D.from_hdf5(
                                        grp['gi_2d'][key]
                                    )
                            self.poni = PONI.from_dict(
                                utils.h5_to_dict(grp['poni'])
                            )

                            self.integrator = self.setup_integrator()

    # ------------------------------------------------------------------
    # NeXus save/load — Phase 2 replacement for save_to_h5/load_from_h5
    # ------------------------------------------------------------------

    def save_to_nexus(self, parent_grp, global_mask=None):
        """Save integration results to a NeXus-formatted HDF5 group.

        Writes into ``parent_grp/<idx>`` (1D) and ``parent_grp/<idx>_2d``
        (2D, optional).  Stores a downsampled thumbnail of the raw image
        in ``parent_grp/<idx>_thumb`` for fast GUI preview.  The thumbnail
        has both arch-level and global masks baked in as NaN pixels.

        Does NOT store full-resolution raw images (map_raw, bg_raw, mask).
        Uses gzip compression only — never lzf (crashes on ARM64 macOS).

        args:
            parent_grp: h5py.Group — typically ``entry/frames/``.
            global_mask: array-like or None — flat indices of detector-level
                masked pixels (from EwaldSphere.global_mask).
        """
        key = f"{self.idx:04d}"

        # ── 1D result ──
        if self.int_1d is not None:
            if key in parent_grp:
                del parent_grp[key]
            grp_1d = parent_grp.create_group(key)
            self.int_1d.to_nexus(grp_1d)
            grp_1d.attrs["source_file"] = self.source_file or ""
            grp_1d.attrs["frame_index"] = int(self.idx)

        # ── Thumbnail (downsampled raw image with mask baked in) ──
        thumb = self.thumbnail
        if thumb is None and self.map_raw is not None:
            try:
                corrected = np.asarray(self.map_raw, dtype=np.float32) - np.asarray(self.bg_raw, dtype=np.float32)
                thumb = _make_thumbnail(corrected, mask_idx=self.mask,
                                        global_mask=global_mask)
            except Exception:
                thumb = None
        if thumb is not None:
            key_thumb = f"{self.idx:04d}_thumb"
            if key_thumb in parent_grp:
                del parent_grp[key_thumb]
            parent_grp.create_dataset(
                key_thumb, data=thumb,
                compression="gzip", compression_opts=1,
            )

        # ── 2D result ──
        if self.int_2d is not None:
            key_2d = f"{self.idx:04d}_2d"
            if key_2d in parent_grp:
                del parent_grp[key_2d]
            grp_2d = parent_grp.create_group(key_2d)
            self.int_2d.to_nexus(grp_2d)
            grp_2d.attrs["source_file"] = self.source_file or ""
            grp_2d.attrs["frame_index"] = int(self.idx)

        # ── GI 1D results ──
        if self.gi_1d:
            for gi_key, res in self.gi_1d.items():
                gi_name = f"{self.idx:04d}_gi1d_{gi_key}"
                if gi_name in parent_grp:
                    del parent_grp[gi_name]
                gi_grp = parent_grp.create_group(gi_name)
                res.to_nexus(gi_grp)
                gi_grp.attrs["source_file"] = self.source_file or ""
                gi_grp.attrs["frame_index"] = int(self.idx)

        # ── GI 2D results ──
        if self.gi_2d:
            for gi_key, res in self.gi_2d.items():
                gi_name = f"{self.idx:04d}_gi2d_{gi_key}"
                if gi_name in parent_grp:
                    del parent_grp[gi_name]
                gi_grp = parent_grp.create_group(gi_name)
                res.to_nexus(gi_grp)
                gi_grp.attrs["source_file"] = self.source_file or ""
                gi_grp.attrs["frame_index"] = int(self.idx)

    def load_from_nexus(self, parent_grp, load_2d=True):
        """Load integration results from a NeXus-formatted HDF5 group.

        Reads from ``parent_grp/<idx>`` (1D), ``parent_grp/<idx>_2d``
        (2D, optional), and ``parent_grp/<idx>_thumb`` (thumbnail).

        args:
            parent_grp: h5py.Group — typically ``entry/frames/``.
            load_2d: bool — if False, skip 2D data and thumbnail.
        """
        key = f"{self.idx:04d}"
        key_2d = f"{self.idx:04d}_2d"
        key_thumb = f"{self.idx:04d}_thumb"

        if key in parent_grp:
            grp_1d = parent_grp[key]
            self.int_1d = IntegrationResult1D.from_hdf5(grp_1d)
            self.source_file = str(grp_1d.attrs.get("source_file", ""))

        # Always load thumbnail (small, useful for preview even in 1D mode)
        if key_thumb in parent_grp:
            self.thumbnail = np.asarray(parent_grp[key_thumb])

        if load_2d and key_2d in parent_grp:
            grp_2d = parent_grp[key_2d]
            self.int_2d = IntegrationResult2D.from_hdf5(grp_2d)

        # GI 1D results
        gi1d_prefix = f"{self.idx:04d}_gi1d_"
        for name in parent_grp:
            if name.startswith(gi1d_prefix):
                gi_key = name[len(gi1d_prefix):]
                self.gi_1d[gi_key] = IntegrationResult1D.from_hdf5(parent_grp[name])

        # GI 2D results
        if load_2d:
            gi2d_prefix = f"{self.idx:04d}_gi2d_"
            for name in parent_grp:
                if name.startswith(gi2d_prefix):
                    gi_key = name[len(gi2d_prefix):]
                    self.gi_2d[gi_key] = IntegrationResult2D.from_hdf5(parent_grp[name])

    def copy(self, include_2d=True):
        """Returns a copy of self.
        """
        arch_copy = EwaldArch(
            copy.deepcopy(self.idx), None,
            copy.deepcopy(self.poni), None,
            copy.deepcopy(self.scan_info), copy.deepcopy(self.ai_args),
            self.file_lock,
            static=copy.deepcopy(self.static), gi=copy.deepcopy(self.gi),
            th_mtr=copy.deepcopy(self.th_mtr),
            sample_orientation=self.sample_orientation,
            series_average=copy.deepcopy(self.series_average)
        )
        arch_copy.integrator = copy.deepcopy(self.integrator)
        arch_copy.arch_lock = Condition()
        arch_copy.int_1d = copy.deepcopy(self.int_1d)
        arch_copy.gi_1d = copy.deepcopy(self.gi_1d)
        arch_copy.source_file = self.source_file
        # Always copy thumbnail — it's small and needed for image preview
        arch_copy.thumbnail = copy.deepcopy(self.thumbnail)
        if include_2d:
            arch_copy.map_raw = copy.deepcopy(self.map_raw)
            arch_copy.bg_raw = copy.deepcopy(self.bg_raw)
            arch_copy.mask = copy.deepcopy(self.mask)
            arch_copy.map_norm = copy.deepcopy(self.map_norm)
            arch_copy.int_2d = copy.deepcopy(self.int_2d)
            arch_copy.gi_2d = copy.deepcopy(self.gi_2d)

        return arch_copy



from __future__ import print_function
from collections import OrderedDict
import copy
from threading import Condition

class Operation(object):
    """Class template for implementing operations
    This is a hold over from paws, will be phased out but not critical.
    """
    # TODO: Remove dependancy on operation

    def __init__(self,inputs,outputs):
        self.default_inputs = copy.deepcopy(inputs)
        self.default_outputs = copy.deepcopy(outputs)
        self.inputs = copy.deepcopy(inputs)
        self.outputs = copy.deepcopy(outputs)
        self.message_callback = self.tagged_print 
        self.data_callback = None 

        # this lock and flag is used to stop long-running Operations.
        # another process can obtain the lock and set the stop_flag,
        # and then the long-running operation should check the flag's value
        # to decide whether or not to stop.
        self.stop_lock = Condition()
        self.stop_flag = False

        # TODO: deprecate instance-level io docs
        self.input_doc = OrderedDict.fromkeys(self.inputs.keys())
        self.output_doc = OrderedDict.fromkeys(self.outputs.keys())

    def __getitem__(self,key):
        return self.__dict__[key]

    def __setitem__(self,key,data):
        self.__dict__[key] = data

    def keys(self):
        #return self.__dict__.keys()
        return['inputs','outputs']

    def tagged_print(self,msg):
        print('[{}] {}'.format(type(self).__name__,msg))

    def run_with(self,**kwargs):
        """Run the Operation with inputs specified by keyword arguments.

        When called, this method loads the default inputs and outputs into the Operation,
        then updates the inputs with the keyword arguments,
        then returns a copy of the return value of the Operation's run() function
        (often, the return value of run() will be the Operation outputs).
        """
        for k in kwargs.keys():
            if not k in self.inputs:
                raise ValueError('Input {} is not valid for Operation {}'.format(k,type(self).__name__))
        self.inputs = copy.deepcopy(self.default_inputs)
        self.outputs = copy.deepcopy(self.default_outputs)
        self.inputs.update(kwargs)
        return copy.deepcopy(self.run()) 

    def run(self):
        """Run the Operation.

        All input data should be specified in Operation.inputs.
        All relevant results should be stored in Operation.outputs.
        """
        pass

    def stop(self):
        with self.stop_lock:
            self.stop_flag = True

    def build_clone(self):
        """Clone the Operation"""
        new_op = self.clone()
        new_op.inputs = copy.deepcopy(self.inputs)
        new_op.outputs = copy.deepcopy(self.outputs)
        return new_op

    @classmethod
    def clone(cls):
        # NOTE: this assumes the subclass constructor 
        # will have default arguments available
        return cls()

    def set_input(self,input_name,val):
        self.inputs[input_name] = val    

    def description(self):
        """Provide a string describing the Operation."""
        return type(self).__name__+": an Operation for PAWS"

